/**********************************************************************
 * gboot - gboot_main.c
 * A simple USB HID bootloader for silicon laboratories USB mcus
 *
 * Copyright (c) 2010 Gabriele Gorla
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **********************************************************************/

#include "gboot_usb_interface.h"
#include "gboot.h"
#include "gboot_usb_core.h"

#pragma codeseg MAINSEG

extern unsigned char ControlReg;
extern setup_packet_t SETUP;
extern unsigned int READ_PTR;

extern __bit rt_ok;

void Handle_Request(void);
void Usb_Reset(void);
void Handle_Control (void);

#ifndef HW_PID
#define HW_PID 0xFFFF
#define HW_REV 0xFF
#endif

#define GBOOT_EXVER (GBOOT_VERSION | (GBOOT_API<<12) )

// pass defines to the assembler/linker
__code  __at(GBOOT_EXVER) unsigned char gboot_version [];

__code  __at(FLASH_START) unsigned char user_code_start [];
__code  __at(FLASH_END) unsigned char user_flash_end [];

__code  __at(HW_PID)  unsigned char hw_pid [];
__code  __at(HW_REV) unsigned char hw_rev [];

// Last packet received from host
__data unsigned char OUT_PACKET[CMD_PACKET_SIZE];

//#ifndef GBOOT_TINY
#if 0
// Next packet to send to host
__data unsigned char IN_PACKET[CMD_PACKET_SIZE];
#endif

void gboot_main(void)
{
  unsigned char i;

  // setup the stack for the C runtime
  __asm
    mov sp,#__start__stack - 1
  __endasm;

  PCA0MD &= ~0x40;                    // Disable Watchdog timer

//-----------------------------------------------------------------------------
// Start firmware unless user requested to enter the bootloader
//-----------------------------------------------------------------------------
    
#ifndef GBOOT_TINY
//#if 1

  // 3 ways to enter bootloader mode
  // 1- shorting P3.0 to gnd
  // 2- corrupted reset vector in firmware
  // 3- sw initiated reboot with xdata flag  
  if( 
     ( (P3_0==1) && 
       (*((__code unsigned char * )FLASH_START)==0x02) ) &&
	 ( !( (RSTSRC&0x10) && (*((__xdata unsigned char * )0x0000)==0xa5) ) )
      ) {

    //          __asm
    //          ljmp	# FLASH_START
    //        __endasm;
    // the above is more correct as the c code will generate
    // a lcall instead of ljmp but the compiler will not
    // optimize the if statement resulting in larger code

    ((void (__code *) (void)) FLASH_START)();
  }

  // reset the bootloader entry flag
  *((__xdata unsigned char * )0x0000)=0x00;
#else 
  // GBOOT_TINY has only 2 ways to enter bootloader mode:
  // 1- shorting P3.0 to gnd
  // 2- sw reset (removed as it does not seem to be reliable on first power on)
  // 3- corrupted reset vector in firmware
#if 0
  if( 
     ( (P3_0==1) && 
       //	 ( !( RSTSRC&0x04) ) &&
       (*((__code unsigned char * )FLASH_START)==0x02) )

      ) {
    ((void (__code *) (void)) FLASH_START)();
  }
#endif
 __asm
    jnb	_P3_0,00102$
    mov	dptr,# FLASH_START
    clr	a
    movc	a,@a+dptr
    mov	r2,a
    cjne	r2,#0x02,00102$
    ljmp	# FLASH_START
00102$:
  __endasm;

#endif
  
//-----------------------------------------------------------------------------
// Sysclk_Init
//-----------------------------------------------------------------------------


  OSCICN |= 0x03;                 // Configure internal oscillator for
                                  // its maximum frequency and enable
                                  // missing clock detector

//      CLKMUL  = 0x00;           // Select internal oscillator as
//      CLKMUL |= 0x80;           // Enable clock multiplier
                                  // input to clock multiplier

  CLKMUL = 0x80;                  // Enable clock multiplier
  CLKMUL |= 0xC0;                 // Initialize the clock multiplier
  // if the above is combined sw reboot occasionally does not work
  //    CLKMUL = 0xC0;                     // Initialize the clock multiplier

  VDM0CN=0x80; // enable VDDMON
  // the above isrequired for flashing to work
  
  // Delay for clock multiplier to begin
  for(i = 0;i < 250; i++) {
    __asm
       nop
    __endasm;
  }

  while(!(CLKMUL & 0x20));            // Wait for multiplier to lock
  //      CLKSEL  = SYS_4X_DIV_2;              // Select 24MHZ system clock
   //     CLKSEL |= USB_4X_CLOCK;             // Select 48MHz USB clock

  CLKSEL  = SYS_4X_DIV_2 | USB_4X_CLOCK;             // Select 48MHz USB clock



  //  VDM0CN=0x80; // enable VDDMON - moved up
  RSTSRC=0x02; // enable reset source
  // the above is also required for flashing to work
  
//-----------------------------------------------------------------------------
// USB_Init
//-----------------------------------------------------------------------------

  USB_WRITE_REG(POWER,  0x08);      // Force Asynchronous USB Reset
  
  USB0XCN = 0xE0;                   // Enable transceiver; select full speed
  USB_WRITE_REG(CLKREC, 0x80);      // Enable clock recovery, single-step
                                    // mode disabled
  
  __asm
    acall _Usb_Reset
  __endasm;
  //  Usb_Reset();

//-----------------------------------------------------------------------------
// Main Loop
//-----------------------------------------------------------------------------
  while (1) {
    unsigned char bCommon, bIn;
    bCommon=USB_READ_REG (CMINT);    // Read all interrupt registers
    bIn=USB_READ_REG (IN1INT);       // this read also clears the register
    
    if (bCommon & rbRSTINT) {         // Handle Reset interrupt
      Usb_Reset();
    }
    if (bIn & rbEP0) {                // Handle SETUP packet received
                                      // or packet transmitted if Endpoint 0
      Handle_Control();             // is in transmit mode
    }
    //    if (bIn & rbIN1) {                // Handle In Packet sent, put new data
                                      // on endpoint 1 fifo
      //	EP_STATUS[1] = EP_IDLE;
    //	EP1_STATUS = 0;
    //      }
  }
}

void Usb_Reset (void)
{
   USB0_STATE = DEV_DEFAULT;           // Set device state to default

   USB_WRITE_REG (POWER, 0x01);      // Clear usb inhibit bit to enable USB
                                       // suspend detection

   EP0_STATUS = EP_IDLE;             // Set default Endpoint Status
   //   EP_STATUS[1] = EP_HALT;
   EP1_STATUS = 1;

   USB_WRITE_INDEX0();         // Set Index to Endpoint Zero
}


void Handle_Control (void)
{
  USB_READ_E0CSR();
  
  if (EP0_STATUS == EP_ADDRESS) {    // Handle Status Phase of Set Address
                                       // command
    USB_WRITE_REG (FADDR, SETUP.wValue.c[LSB]);
    EP0_STATUS = EP_IDLE;
  }
  
  if (ControlReg & rbSTSTL) {          // If last packet was a sent stall,
                                       // reset STSTL bit and return EP0
                                       // to idle state
    USB_WRITE_E0CSR(0);
    EP0_STATUS = EP_IDLE;
    return;
  }
  
  if (ControlReg & rbSUEND) {      // If last SETUP transaction was
                                   // ended prematurely then set
    //    USB_WRITE_E0CSR(rbDATAEND);    // Serviced SETUP End bit and return EP0
    //    USB_WRITE_E0CSR(rbSSUEND);
    USB_WRITE_E0CSR(rbSSUEND|rbDATAEND);
    EP0_STATUS = EP_IDLE;          // To idle state
  }

  if (EP0_STATUS == EP_IDLE) {     // If Endpoint 0 is in idle mode
    if (ControlReg & rbOPRDY)  {   // Make sure that EP 0 has an Out Packet
                                   // ready from host although if EP0
                                   // is idle, this should always be the
                                   // case
      Fifo_Read (FIFO_EP0, sizeof(setup_packet_t) , (unsigned char __data *)&SETUP);
      // Get SETUP Packet off of Fifo,
      Handle_Request();   
    }
  }
  
  if (EP0_STATUS == EP_TX) {         // See if endpoint should transmit
    if (!(ControlReg & rbINPRDY) ) {  // Don't overwrite last packet
      // Read control register
      //	ControlReg=USB_READ_BYTE (E0CSR);
      USB_READ_E0CSR();
      
      // Check to see if SETUP End or Out Packet received, if so do not put
      // any new data on FIFO
      // if ((!(ControlReg & rbSUEND)) || (!(ControlReg & rbOPRDY)))
      if (!(ControlReg & (rbSUEND | rbOPRDY))) {
	unsigned char i;
	
      while (USB0ADR & 0x80);       // Wait for BUSY->'0' (data ready)
	//	USB_POLL();
	USB0ADR = FIFO_EP0;           // Set address (mask out bits7-6)
	
	// Write <NumBytes> to the selected FIFO
	for (i=0; i<DATASIZE; i++) {
	  USB0DAT = TXDATAPTR[i];
	  //	  USB_POLL();
      while (USB0ADR & 0x80);       // Wait for BUSY->'0' (data ready)
	}
	EP0_STATUS = EP_IDLE; // Return EP 0 to idle state
	USB_WRITE_E0CSR( rbINPRDY|rbDATAEND);
      }
    }
  }
  
  if (EP0_STATUS == EP_RX) {     // See if endpoint should receive
    // Read control register
    USB_READ_E0CSR();
    
    if (ControlReg & rbOPRDY) {        // Verify packet was received
      Fifo_Read (FIFO_EP0, DATASIZE, (unsigned char __data *) RXDATAPTR);
      
      EP0_STATUS = EP_IDLE;    // set Endpoint to IDLE
      if ( (SETUP.bRequest == SET_REPORT) ) {
	ReportHandler();
      }
      
      USB_WRITE_E0CSR_END();
    }
  }
}

void Fifo_Read (unsigned char addr, unsigned char uNumBytes,
		unsigned char __data * pData) 
{
  unsigned char i;
  
  if (uNumBytes) {                     // Check if >0 bytes requested,
    USB0ADR = (addr);                // Set address
    USB0ADR |= 0xC0;                 // Set auto-read and initiate
    // first read
    
    // Unload <NumBytes> from the selected FIFO
    for(i=0;i< (uNumBytes);i++) {
      //USB_POLL();
      while (USB0ADR & 0x80);       // Wait for BUSY->'0' (data ready)
      pData[i] = USB0DAT;           // Copy data byte
    }
    
    //while(USB0ADR & 0x80);         // Wait for BUSY->'0' (data ready)
    USB0ADR = 0;                     // Clear auto-read
  }
}


unsigned char USB_READ_REG(unsigned char a)
{
  //USB_POLL();
  while(USB0ADR & 0x80); 
  USB0ADR = (0x80|a);
  //  USB_POLL();
    while(USB0ADR & 0x80); 
  return USB0DAT;
}

void USB_READ_E0CSR()
{
  //USB_POLL();
    while(USB0ADR & 0x80); 
  USB0ADR = (0x80|E0CSR);
  //USB_POLL(); 
    while(USB0ADR & 0x80); 
  ControlReg = USB0DAT;
}
